"""Entry point of the cloai package."""
import asyncio
import os
import sys


def main() -> None:
    """Entry point for the CLI."""
    if "OPENAI_API_KEY" not in os.environ:
        sys.stderr.write("Error: Please set the OPENAI_API_KEY environment variable.")
        sys.exit(1)

    from cloai.cli import parser

    asyncio.run(parser.parse_args())


if __name__ == "__main__":
    main()



"""Contains the core business logic of the OpenAI CLI."""
import asyncio
import logging
import pathlib
import tempfile
from typing import Literal

import aiofiles
import ffmpeg
import yaml

from cloai import openai_api
from cloai.core import config, exceptions, utils

settings = config.get_settings()
logger = logging.getLogger(settings.LOGGER_NAME)
PROMPT_FILE = settings.PROMPT_FILE
MAX_FILE_SIZE = 24_500_000  # Max size is 25MB, but we leave some room for error.


class ChatCompletion:
    """A class for running the Chat Completion model.

    Attributes:
        user_prompt: The prompt to use.
        system_prompt: The system prompt to use.
        model: The model to use for chat completion.
    """

    def __init__(  # noqa: PLR0913
        self,
        model: Literal["gpt-4", "gpt-3.5-turbo", "gpt-4-1106-preview"],
        *,
        user_prompt: str | None = None,
        user_prompt_file: pathlib.Path | None = None,
        system_prompt: str | None = None,
        system_prompt_file: pathlib.Path | None = None,
        system_preset: str | None = None,
    ) -> None:
        """Initializes a ChatCompletion object.

        Args:
            user_prompt: The prompt to use.
            user_prompt_file: The file containing the prompt to use.
            system_prompt: The system prompt to use.
            system_prompt_file: The file containing the system prompt to use.
            system_preset: The preset system prompt to use.
            model: The model to use for chat completion.

        """
        logger.debug("Initializing ChatCompletion.")
        self._validate_initialization(
            user_prompt,
            user_prompt_file,
            system_prompt,
            system_prompt_file,
            system_preset,
        )
        self.user_prompt: str = (
            self._read_file(user_prompt_file) if user_prompt_file else user_prompt  # type: ignore[assignment]
        )
        self.system_prompt = self._determine_system_prompt(
            system_prompt,
            system_prompt_file,
            system_preset,
        )
        self.model = model

    async def run(self) -> str:
        """Runs the chat completion.

        Returns:
            The completed chat response as a string.
        """
        logger.debug("Running chat completion.")
        return await openai_api.ChatCompletion().run(
            model=self.model,
            user_prompt=self.user_prompt,
            system_prompt=self.system_prompt,
        )

    @staticmethod
    def _validate_initialization(
        user_prompt: str | None = None,
        user_prompt_file: pathlib.Path | None = None,
        system_prompt: str | None = None,
        system_prompt_file: pathlib.Path | None = None,
        system_preset: str | None = None,
    ) -> None:
        """Validates the arguments passed to the constructor.

        Args:
            user_prompt: The prompt to use.
            user_prompt_file: The file containing the prompt to use.
            system_prompt: The system prompt to use.
            system_prompt_file: The file containing the system prompt to use.
            system_preset: The preset system prompt to use.

        Raises:
            ValueError: If prompt and prompt_file are both None.
            ValueError: If not just one of system_prompt, system_prompt_file,
                and system_preset are provided.
        """
        user_args = (user_prompt, user_prompt_file)
        if sum(arg is not None for arg in user_args) != 1:
            msg = "Either prompt or prompt_file must be provided."
            raise exceptions.LoggedValueError(msg)

        system_args = (system_prompt, system_prompt_file, system_preset)
        if sum(arg is not None for arg in system_args) != 1:
            msg = (
                "Exactly one of system_prompt, system_prompt_file,"
                "or system_preset must be provided."
            )
            raise exceptions.LoggedValueError(msg)

    @staticmethod
    def _read_file(file_path: pathlib.Path) -> str:
        """Reads the contents of a file.

        Args:
            file_path: The path to the file.

        Returns:
            The contents of the file as a string.
        """
        if file_path.suffix == ".pdf":
            return utils.pdf_to_str(file_path)
        if file_path.suffix == ".docx":
            return utils.docx_to_str(file_path)
        if file_path.suffix != ".txt":
            logger.warning(
                "File %s has an unsupported extension, treating it as a .txt file.",
                file_path,
            )
        return utils.txt_to_str(file_path)

    def _determine_system_prompt(
        self,
        system_prompt: str | None = None,
        system_prompt_file: pathlib.Path | None = None,
        system_preset: str | None = None,
    ) -> str:
        """Determines the system prompt to use.

        Args:
            system_prompt: The system prompt to use.
            system_prompt_file: The file containing the system prompt to use.
            system_preset: The preset system prompt to use.

        Returns:
            The determined system prompt as a string.
        """
        if system_prompt_file:
            return self._read_file(system_prompt_file)
        if system_preset:
            with PROMPT_FILE.open() as file:
                prompts = yaml.safe_load(file)
                return prompts["system"][system_preset]
        if system_prompt:
            return system_prompt
        msg = "No system prompt provided."
        raise exceptions.LoggedValueError(msg)


async def speech_to_text(
    filename: pathlib.Path,
    model: str,
    *,
    clip: bool = False,
    language: config.WhisperLanguages = config.WhisperLanguages.ENGLISH,
) -> str:
    """Transcribes audio files with OpenAI's TTS models.

    Args:
        filename: The file to transcribe. Can be any format that ffmpeg supports.
        model: The transcription model to use.
        clip: Whether to clip the file if it is too large, defaults to False.
        language: The language used in the audio file.
    """
    logger.debug("Transcribing audio.")
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = pathlib.Path(temp_dir) / "temp.mp3"
        ffmpeg.input(filename).output(str(temp_file)).overwrite_output().run()

        if clip:
            files = list(utils.clip_audio(temp_file, temp_dir, MAX_FILE_SIZE))
        else:
            files = [temp_file]

        stt = openai_api.SpeechToText()
        transcription_promises = [
            stt.run(filename, model=model, language=language) for filename in files
        ]
        transcriptions = await asyncio.gather(*transcription_promises)

        return " ".join(transcriptions)


async def text_to_speech(
    text: str,
    output_file: str,
    model: Literal["tts-1", "tts-1-hd"],
    voice: Literal["alloy", "echo", "fable", "onyx", "nova", "shimmer"],
) -> None:
    """Converts text to speech with OpenAI's TTS models.

    Args:
        text: The text to convert to speech.
        output_file: The name of the output file.
        model: The model to use.
        voice: The voice to use.
    """
    logger.debug("Converting text to speech.")
    tts = openai_api.TextToSpeech()
    audio_bytes = await tts.run(text, model=model, voice=voice)
    await utils.save_file(output_file, audio_bytes)


async def image_generation(  # noqa: PLR0913
    prompt: str,
    output_base_name: str,
    model: str,
    size: Literal["256x256", "512x512", "1024x1024", "1792x1024", "1024x1792"] | None,
    quality: Literal["standard", "hd"],
    n: int,
) -> None:
    """Generates an image from text with OpenAI's Image Generation models.

    Args:
        prompt: The text to generate an image from.
        output_base_name: The base name of the output file.
        model: The model to use.
        size: The size of the generated image.
        quality: The quality of the generated
        image. Defaults to "standard".
        n: The number of images to generate.

    Returns:
        bytes: The generated image as bytes.

    Notes:
        At present, the image generation API of dalle-3 only supports generating
        one image at a time. Instead, we call the API once for each image we want
        to generate.
    """
    logger.debug("Generating %s images.", n)
    image_generation = openai_api.ImageGeneration()
    url_promises = [
        image_generation.run(
            prompt,
            model=model,
            size=size,
            quality=quality,
            n=1,
        )
        for _ in range(n)
    ]
    urls = [url[0] for url in await asyncio.gather(*url_promises)]
    if all(url is None for url in urls):
        msg = "No images were generated."
        raise exceptions.OpenAIError(msg)
    urls_not_none = [url for url in urls if url is not None]

    await asyncio.gather(
        *[
            utils.download_file(f"{output_base_name}_{index}.png", url)
            for index, url in enumerate(urls_not_none)
        ],
    )


async def get_embedding(
    text_file: pathlib.Path,
    output_file: pathlib.Path,
    model: Literal[
        "text-embedding-3-small",
        "text-embedding-3-large",
    ] = "text-embedding-3-large",
    *,
    keep_new_lines: bool = False,
) -> None:
    """Get the embedding using OpenAI's Embedding models.

    Args:
        text_file: the text file to embed.
        model: the name of the Embedding model to use,
        defaults to text-embedding-3-large.
        output_file: the name of the CSV output file.
        keep_new_lines: Whether to keep or remove line breaks,
        defaults to False.
    """
    async with aiofiles.open(text_file, mode="r") as file:
        text = await file.read()

    get_embedding = openai_api.Embedding()

    embedding = await get_embedding.run(
        text=text,
        model=model,
        keep_new_lines=keep_new_lines,
    )
    await utils.save_csv(output_file, embedding)



"""Configuration for the cloai module."""
import enum
import functools
import logging
import pathlib
from importlib import metadata
from typing import Literal

import pydantic
import pydantic_settings


def get_version() -> str:
    """Return the version of the package."""
    try:
        return metadata.version("cloai")
    except metadata.PackageNotFoundError:
        return "unknown"


class WhisperLanguages(str, enum.Enum):
    """The languages for the whisper model."""

    AFRIKAANS = "af"
    ARABIC = "ar"
    ARMENIAN = "hy"
    AZERBAIJANI = "az"
    BELARUSIAN = "be"
    BOSNIAN = "bs"
    BULGARIAN = "bg"
    CATALAN = "ca"
    CHINESE = "zh"
    CROATIAN = "hr"
    CZECH = "cs"
    DANISH = "da"
    DUTCH = "nl"
    ENGLISH = "en"
    ESTONIAN = "et"
    FINNISH = "fi"
    FRENCH = "fr"
    GALICIAN = "gl"
    GERMAN = "de"
    GREEK = "el"
    HEBREW = "he"
    HINDI = "hi"
    HUNGARIAN = "hu"
    ICELANDIC = "is"
    INDONESIAN = "id"
    ITALIAN = "it"
    JAPANESE = "ja"
    KANNADA = "kn"
    KAZAKH = "kk"
    KOREAN = "ko"
    LATVIAN = "lv"
    LITHUANIAN = "lt"
    MACEDONIAN = "mk"
    MALAY = "ms"
    MARATHI = "mr"
    MAORI = "mi"
    NEPALI = "ne"
    NORWEGIAN = "no"
    PERSIAN = "fa"
    POLISH = "pl"
    PORTUGUESE = "pt"
    ROMANIAN = "ro"
    RUSSIAN = "ru"
    SERBIAN = "sr"
    SLOVAK = "sk"
    SLOVENIAN = "sl"
    SPANISH = "es"
    SWAHILI = "sw"
    SWEDISH = "sv"
    TAGALOG = "tl"
    TAMIL = "ta"
    THAI = "th"
    TURKISH = "tr"
    UKRAINIAN = "uk"
    URDU = "ur"
    VIETNAMESE = "vi"
    WELSH = "cy"


class Settings(pydantic_settings.BaseSettings):
    """Represents the settings for the cloai module."""

    LOGGER_NAME: str = "cloai"
    LOGGER_VERBOSITY: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO"

    OPENAI_API_KEY: pydantic.SecretStr | None = pydantic.Field(
        None,
        json_schema_extra={
            "env": "OPENAI_API_KEY",
            "description": "The API key for OpenAI.",
        },
    )

    PROMPT_FILE: pydantic.FilePath = pydantic.Field(
        pathlib.Path(__file__).parent.parent / "prompts.yaml",
    )


@functools.lru_cache
def get_settings() -> Settings:
    """Cached fetcher for the API settings.

    Returns:
        The settings for the API.
    """
    return Settings()  # type: ignore[call-arg]


def initialize_logger() -> logging.Logger:
    """Initializes the logger for the API."""
    settings = get_settings()
    logger = logging.getLogger(settings.LOGGER_NAME)
    if settings.LOGGER_VERBOSITY is not None:
        logger.setLevel(settings.LOGGER_VERBOSITY)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(filename)s:%(lineno)s - %(funcName)s - %(message)s",  # noqa: E501
    )

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


def get_logger() -> logging.Logger:
    """Gets the logger for the API.

    Returns:
        The logger for the API.
    """
    settings = get_settings()
    logger = logging.getLogger(settings.LOGGER_NAME)
    if logger.hasHandlers():
        return logger
    return initialize_logger()



"""Command line interface for the OpenAI API."""
import argparse
import pathlib
import sys

from cloai.cli import commands
from cloai.core import config, exceptions

logger = config.get_logger()


PARSER_DEFAULTS = {
    "epilog": "Please report issues at https://github.com/childmindresearch/cloai.",
    "formatter_class": argparse.ArgumentDefaultsHelpFormatter,
}


async def parse_args() -> None:
    """Parse command line arguments and execute the corresponding command."""
    parser = create_parser()
    validated_args = get_args(parser)
    result = await run_command(validated_args)
    if isinstance(result, str):
        sys.stdout.write(result)


def get_args(parser: argparse.ArgumentParser) -> argparse.Namespace:
    """Parse the command line arguments using the provided ArgumentParser object.

    Args:
        parser: The ArgumentParser object to parse the arguments.

    Returns:
        argparse.Namespace: The parsed command line arguments.

    Raises:
        SystemExit: If the command is not specified.
    """
    args = parser.parse_args()
    if args.command is None:
        parser.print_usage()
        sys.exit(1)
    return _arg_validation(args)


def create_parser() -> argparse.ArgumentParser:
    """Create and configure the argument parser for the CLI.

    Returns:
        argparse.ArgumentParser: The configured argument parser.
    """
    parser = argparse.ArgumentParser(
        prog="cloai",
        description="""
        CLI wrapper for OpenAI's API. All commands require the OPENAI_API_KEY
        environment variable to be set to a valid OpenAI API key.""",
        **PARSER_DEFAULTS,  # type: ignore[arg-type]
    )

    version = config.get_version()
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {version}",
    )

    subparsers = parser.add_subparsers(dest="command")
    _add_chat_completion_parser(subparsers)
    _add_stt_parser(subparsers)
    _add_tts_parser(subparsers)
    _add_image_generation_parser(subparsers)
    _add_embedding_parser(subparsers)
    return parser


async def run_command(args: argparse.Namespace) -> str | bytes | None:
    """Executes the specified command based on the provided arguments.

    Args:
        args: The parsed command-line arguments.

    Returns:
        str, bytes, None: The result of the executed command.
    """
    if getattr(args, "command", None) is None:
        msg = "No command provided."
        raise exceptions.InvalidArgumentError(msg)

    if args.command == "dalle":
        await commands.image_generation(
            prompt=args.prompt,
            output_base_name=args.base_image_name,
            model=args.model,
            size=args.size,
            quality=args.quality,
            n=args.number,
        )
        return None
    if args.command == "gpt":
        return await commands.ChatCompletion(
            user_prompt=args.user_prompt,
            user_prompt_file=args.user_prompt_file,
            system_prompt=args.system_prompt,
            system_prompt_file=args.system_prompt_file,
            system_preset=args.system_preset,
            model=args.model,
        ).run()
    if args.command == "tts":
        await commands.text_to_speech(
            text=args.text,
            model=args.model,
            voice=args.voice,
            output_file=args.output_file,
        )
        return None
    if args.command == "whisper":
        return await commands.speech_to_text(
            filename=args.filename,
            model=args.model,
            clip=args.clip,
            language=config.WhisperLanguages[args.language],
        )
    if args.command == "embedding":
        await commands.get_embedding(
            text_file=args.text_file,
            output_file=args.output_file,
            model=args.model,
            keep_new_lines=args.keep_new_lines,
        )
        return None
    msg = f"Unknown command {args.command}."
    raise exceptions.InvalidArgumentError(msg)


def _add_chat_completion_parser(
    subparsers: argparse._SubParsersAction,
) -> None:
    """Get the argument parser for the 'gpt' command.

    Args:
        subparsers: The subparsers object to add the 'gpt' command to.

    Returns:
        argparse.ArgumentParser: The argument parser for the 'gpt' command.
    """
    chat_parser = subparsers.add_parser(
        "gpt",
        description="Completes text with OpenAI's GPT models.",
        help="Completes text with OpenAI's GPT models.",
        **PARSER_DEFAULTS,  # type: ignore[arg-type]
    )

    user_group = chat_parser.add_argument_group(
        "User Prompts",
        """The prompts provided by the user. One must be provided and these
        arguments are mutually exclusive.""",
    )
    user_group_exclusive = user_group.add_mutually_exclusive_group(required=True)
    user_group_exclusive.add_argument(
        "-p",
        "--user-prompt",
        help="The user prompt to complete.",
        type=str,
    )
    user_group_exclusive.add_argument(
        "-f",
        "--user-prompt-file",
        help="The file containing the prompt to complete.",
        type=pathlib.Path,
    )

    system_group = chat_parser.add_argument_group(
        "System Prompts",
        """The prompts povided to the system. One must be provided and these
        arguments are mutually exclusive.""",
    )
    system_group_exclusive = system_group.add_mutually_exclusive_group(required=True)
    system_group_exclusive.add_argument(
        "-s",
        "--system-prompt",
        help="The system prompt.",
    )
    system_group_exclusive.add_argument(
        "-t",
        "--system-prompt-file",
        help="The file containing the system prompt.",
        type=pathlib.Path,
    )
    system_group_exclusive.add_argument(
        "-r",
        "--system-preset",
        help="""The preset system prompt. The following prompts are available:
        1) "summary": Summarizes the user prompt; 2) "transcript": Edits a
         audio transcript for coherence; 3) "simplify": Simplifies the reading
         level of the user prompt.""",
        choices=("summary", "transcript", "simplify"),
    )

    chat_parser.add_argument(
        "-m",
        "--model",
        help="The model to use.",
        choices=("gpt-4", "gpt-3.5-turbo", "gpt-4-1106-preview"),
        default="gpt-4",
    )


def _add_stt_parser(
    subparsers: argparse._SubParsersAction,
) -> None:
    """Get the argument parser for the "whisper" command.

    Args:
        subparsers: The subparsers object to add the
            "whisper" command to.

    Returns:
        argparse.ArgumentParser: The argument parser for the "whisper" command.
    """
    stt_parser = subparsers.add_parser(
        "whisper",
        description="Transcribes audio files with OpenAI's STT models.",
        help="Transcribes audio files with OpenAI's STT models.",
        **PARSER_DEFAULTS,  # type: ignore[arg-type]
    )
    stt_parser.add_argument(
        "filename",
        help="The file to transcribe. Can be any format that ffmpeg supports.",
        type=pathlib.Path,
    )
    stt_parser.add_argument(
        "-c",
        "--clip",
        help="Clip the file if it is too large.",
        action="store_true",
    )
    stt_parser.add_argument(
        "-m",
        "--model",
        help=("The transcription model to use."),
        type=lambda x: x.lower(),
        choices=["whisper-1"],
        default="whisper-1",
    )
    stt_parser.add_argument(
        "--language",
        help="The language of the audio file.",
        type=lambda x: x.upper(),
        choices=[language.name for language in config.WhisperLanguages],
        default="ENGLISH",
    )


def _add_tts_parser(
    subparsers: argparse._SubParsersAction,
) -> None:
    """Get the argument parser for the "tts" command.

    Args:
        subparsers (argparse._SubParsersAction): The subparsers object to add the
            "tts" command to.

    Returns:
        argparse.ArgumentParser: The argument parser for the "tts" command.
    """
    tts_parser = subparsers.add_parser(
        "tts",
        description="Generates audio files with OpenAI's Jukebox.",
        help="Generates audio files with OpenAI's TTS models.",
        **PARSER_DEFAULTS,  # type: ignore[arg-type]
    )
    tts_parser.add_argument(
        "text",
        help="The text to generate audio from.",
        type=str,
    )
    tts_parser.add_argument(
        "output_file",
        help="The name of the output file.",
        type=pathlib.Path,
    )
    tts_parser.add_argument(
        "-m",
        "--model",
        help=("The model to use."),
        choices=["tts-1"],
        default="tts-1",
    )
    tts_parser.add_argument(
        "-v",
        "--voice",
        help="The voice to use.",
        choices=["alloy", "echo", "fable", "onyx", "nova", "shimmer"],
        default="onyx",
    )


def _add_image_generation_parser(
    subparsers: argparse._SubParsersAction,
) -> None:
    image_generation_parser = subparsers.add_parser(
        "dalle",
        description="Generates images with OpenAI's DALL-E.",
        help="Generates images with OpenAI's image generation models.",
        **PARSER_DEFAULTS,  # type: ignore[arg-type]
    )
    image_generation_parser.add_argument(
        "prompt",
        help="The prompt to generate images from.",
        type=str,
    )
    image_generation_parser.add_argument(
        "base_image_name",
        help="The base name for output images.",
        type=str,
    )
    image_generation_parser.add_argument(
        "-m",
        "--model",
        help=(
            "The model to use. Consult OpenAI's documentation for an up-to-date list"
            " of models."
        ),
        type=lambda x: x.lower(),
        default="dall-e-3",
    )
    image_generation_parser.add_argument(
        "-s",
        "--size",
        help="The size of the generated image.",
        type=lambda x: x.lower(),
        choices=["256x256", "512x512", "1024x1024", "1792x1024", "1024x1792"],
        default="1024x1024",
    )
    image_generation_parser.add_argument(
        "-q",
        "--quality",
        help="The quality of the generated image.",
        type=lambda x: x.lower(),
        choices=["standard", "hd"],
        default="standard",
    )
    image_generation_parser.add_argument(
        "-n",
        "--number",
        help="The number of images to generate.",
        type=_positive_int,
        default=1,
    )


def _add_embedding_parser(
    subparsers: argparse._SubParsersAction,
) -> None:
    """Get the argument parser for the "embedding" command.

    Args:
        subparsers: The subparsers object to add the "embedding" command to.

    Returns:
        argparse.ArgumentParser: The argument parser for the "embedding" command.
    """
    embedding_parser = subparsers.add_parser(
        "embedding",
        description="Generates embedding with OpenAI's Text Embedding models.",
        help="Generates embedding with OpenAI's Text Embedding models.",
        **PARSER_DEFAULTS,  # type: ignore[arg-type]
    )
    embedding_parser.add_argument(
        "text_file",
        help="The text file to generate an embedding from.",
        type=pathlib.Path,
    )
    embedding_parser.add_argument(
        "output_file",
        help="The name of the CSV output file.",
        type=pathlib.Path,
    )
    embedding_parser.add_argument(
        "-m",
        "--model",
        help=("The model to use."),
        choices=["text-embedding-3-small", "text-embedding-3-large"],
        default="text-embedding-3-large",
    )
    embedding_parser.add_argument(
        "--keep-new-lines",
        dest="keep_new_lines",
        help=("Keeps line breaks in text if specified."),
        action=argparse.BooleanOptionalAction,
        default=False,
    )


def _arg_validation(args: argparse.Namespace) -> argparse.Namespace:
    """Validate the parsed arguments.

    Validation across arguments is not possible with the built-in argparse
    validation. This function performs validation across arguments.

    Args:
        args: The parsed command-line arguments.

    Returns:
        argparse.ArgumentParser: The validated arguments.
    """
    if args.command == "dalle":  # noqa: SIM102 # Allows logical expansion across commands.
        if args.model == "dall-e-3" and args.size in ["256x256", "512x512"]:
            msg = "The dall-e-3 model does not support 256x256 or 512x512 images."
            raise exceptions.InvalidArgumentError(msg)

    return args


def _positive_int(value: int | str) -> int:
    """Ensures the value is a positive integer.

    Args:
        value: The value.

    Returns:
        int: The positive integer.

    Raises:
        exceptions.InvalidArgumentError: If the value is not an integer or not a
        positive integer.
    """
    if int(value) != float(value):
        msg = f"{value} is not an integer."
        raise exceptions.InvalidArgumentError(msg)
    if int(value) <= 0:
        msg = f"{value} is not a positive integer."
        raise exceptions.InvalidArgumentError(msg)
    return int(value)



"""Utility functions for cloai."""
import math
import pathlib
import uuid
from collections.abc import Generator

import aiocsv
import aiofiles
import aiohttp
import docx
import ffmpeg
import pypdf

from cloai.core import config

logger = config.get_logger()


def clip_audio(
    filename: str | pathlib.Path,
    out_dir: str | pathlib.Path,
    target_size: int,
) -> Generator[pathlib.Path, None, None]:
    """Clips the file to the maximum size.

    Args:
    ----
        filename: The file to clip.
        out_dir: The directory to save the clipped files to.
        target_size: The target size of the clipped files.

    """
    logger.warning(
        "File too large. Clipping may lead to inaccurate results around the clips.",
    )

    file_size = get_file_size(filename)
    n_files = (file_size // target_size) + 1
    file_duration = get_audio_duration(filename)
    clip_duration = math.ceil(file_duration / n_files)

    uuid_id = uuid.uuid4()
    output_file_pattern = pathlib.Path(out_dir) / f"audio_{uuid_id}_%03d.mp3"
    stream = (
        ffmpeg.input(str(filename))
        .output(
            str(output_file_pattern),
            f="segment",
            segment_time=clip_duration,
        )
        .overwrite_output()
    )

    logger.warning("Running ffmpeg: %s", " ".join(stream.compile()))
    stream.run()

    files = list(pathlib.Path(out_dir).glob(f"*{uuid_id}*.mp3"))
    files.sort()
    yield from files


async def save_file(filename: str | pathlib.Path, content: bytes) -> None:
    """Saves content to a file asynchronously.

    Args:
    ----
        filename: The name of the file to save the content to.
        content: The content to save to the file.

    """
    async with aiofiles.open(filename, "wb") as file:
        await file.write(content)


async def save_csv(filename: str | pathlib.Path, content: list[float]) -> None:
    """Saves content to a csv file asynchronously.

    Args:
    ----
        filename: The name of the file to save the content to.
        content: The content to save to the file.

    """
    async with aiofiles.open(filename, "w") as file:
        writer = aiocsv.AsyncWriter(file)
        await writer.writerow(content)


async def download_file(filename: str | pathlib.Path, url: str) -> None:
    """Downloads a file from a URL.

    Args:
    ----
        filename: The name of the file to download.
        url: The URL to download the file from.

    """
    async with aiohttp.ClientSession() as session, session.get(url) as response:
        response.raise_for_status()
        async with aiofiles.open(filename, "wb") as file:
            while True:
                chunk = await response.content.read(1024)
                if not chunk:
                    break
                await file.write(chunk)


def get_audio_duration(filename: str | pathlib.Path) -> float:
    """Gets the duration of the audio file.

    Args:
    ----
        filename: The name of the audio file.

    Returns:
    -------
        float: The duration of the audio file.

    """
    probe = ffmpeg.probe(filename)
    return float(probe["format"]["duration"])


def get_file_size(filename: str | pathlib.Path) -> int:
    """Gets the size of the file.

    Args:
    ----
        filename: The name of the file.

    Returns:
    -------
        int: The size of the file.

    """
    return pathlib.Path(filename).stat().st_size


def pdf_to_str(file_path: str | pathlib.Path) -> str:
    """Convert a PDF file to text.

    Args:
    ----
        file_path: The path to the PDF file.

    Returns:
    -------
        str: The extracted text from the PDF file.

    """
    with pathlib.Path(file_path).open("rb") as file:
        reader = pypdf.PdfReader(file)
        return " ".join(
            [
                reader.pages[page_num].extract_text()
                for page_num in range(len(reader.pages))
            ],
        )


def docx_to_str(file_path: str | pathlib.Path) -> str:
    """Convert a docx file to text.

    Args:
    ----
        file_path: The path to the docx file.

    Returns:
    -------
        str: The extracted text from the docx file.

    """
    document = docx.Document(file_path)
    return " ".join([paragraph.text for paragraph in document.paragraphs])


def txt_to_str(file_path: str | pathlib.Path) -> str:
    """Convert a txt file to text.

    Args:
    ----
        file_path: The path to the txt file.

    Returns:
    -------
        str: The extracted text from the txt file.

    """
    with pathlib.Path(file_path).open() as file:
        return file.read()



"""This module contains interactions with OpenAI models."""
import abc
import logging
import pathlib
from typing import Any, Literal, TypedDict, TypeVar

import instructor
import openai
import pydantic

from cloai.core import config, exceptions

settings = config.get_settings()
OPENAI_API_KEY = settings.OPENAI_API_KEY
LOGGER_NAME = settings.LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)

T = TypeVar("T")


class Message(TypedDict):
    """A message object."""

    role: Literal["assistant", "system", "user"]
    content: str


class OpenAIBaseClass(pydantic.BaseModel, abc.ABC):
    """An abstract base class for OpenAI models.

    This class initializes the OpenAI client and requires a run method to be
    implemented.

    Attributes:
        api_key: The OpenAI API key.
        client: The OpenAI client used to interact with the model.
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
    )

    api_key: pydantic.SecretStr | None = OPENAI_API_KEY
    client: openai.AsyncOpenAI = pydantic.Field(init=False, default=None)

    def model_post_init(self, __context: Any) -> None:  # noqa: ANN401
        """Initializes a new instance of the OpenAIBaseClass class.

        Args:
            api_key: The OpenAI API key.
        """
        if self.api_key is None:
            msg = "No OpenAI API key provided."
            raise exceptions.OpenAIError(msg)
        self.client = openai.AsyncOpenAI(api_key=self.api_key.get_secret_value())

    @abc.abstractmethod
    async def run(self, *_args: Any, **_kwargs: Any) -> Any:  # noqa: ANN401
        """Runs the model."""
        ...


class ChatCompletion(OpenAIBaseClass):
    """A class for running the Chat Completion models."""

    async def run(
        self,
        user_prompt: str,
        system_prompt: str,
        model: str = "gpt-4",
    ) -> str:
        """Runs the Chat Completion model.

        Args:
            user_prompt: The user's prompt.
            system_prompt: The system's prompt.
            model: The name of the Chat Completion model to use.

        Returns:
            The model's response.
        """
        system_message = Message(role="system", content=system_prompt)
        user_message = Message(role="user", content=user_prompt)
        response = await self.client.chat.completions.create(
            model=model,
            messages=[system_message, user_message],  # type: ignore[list-item]
        )
        if not response.choices[0].message.content:
            msg = "No response from OpenAI."
            raise exceptions.OpenAIError(msg)
        return response.choices[0].message.content


class ChatCompletionInstructor(OpenAIBaseClass):
    """A class for running the Chat Completion models using the instructor library.

    This class is intended to only be exposed to the user through the Python interface.
    """

    def __init__(self) -> None:
        """Patches the OpenAI library to use instructor."""
        super().__init__()
        self.client = instructor.patch(self.client)

    async def run(  # noqa: PLR0913
        self,
        user_prompt: str,
        system_prompt: str,
        response_model: T,
        model: str = "gpt-4",
        max_retries: int = 1,
    ) -> T:
        """Runs the Chat Completion model.

        Args:
            user_prompt: The user's prompt.
            system_prompt: The system's prompt.
            response_model: The response model to return.
            model: The name of the Chat Completion model to use.
            max_retries: The maximum number of retries to attempt.

        Returns:
            The model's response.
        """
        system_message = Message(role="system", content=system_prompt)
        user_message = Message(role="user", content=user_prompt)
        return await self.client.chat.completions.create(  # type: ignore[call-overload]
            model=model,
            messages=[system_message, user_message],  # type: ignore[list-item]
            response_model=response_model,
            max_retries=max_retries,
        )


class TextToSpeech(OpenAIBaseClass):
    """A class for running the Text-To-Speech models."""

    async def run(
        self,
        text: str,
        model: Literal["tts-1", "tts-1-hd"] = "tts-1",
        voice: Literal["alloy", "echo", "fable", "onyx", "nova", "shimmer"] = "onyx",
    ) -> bytes:
        """Runs the Text-To-Speech model.

        Args:
            text: The text to convert to speech.
            output_file: The name of the output file.
            model: The name of the Text-To-Speech model to use.
            voice: The voice to use.

        Returns:
            The model's response.
        """
        response = await self.client.audio.speech.create(
            model=model,
            voice=voice,
            input=text,
        )
        return response.content


class SpeechToText(OpenAIBaseClass):
    """A class for running the Speech-To-Text models."""

    async def run(
        self,
        audio_file: pathlib.Path | str,
        model: str = "whisper-1",
        language: config.WhisperLanguages | str = config.WhisperLanguages.ENGLISH,
    ) -> str:
        """Runs the Speech-To-Text model.

        Args:
            audio_file: The audio to convert to text.
            model: The name of the Speech-To-Text model to use.
            language: The language of the audio. Can be both provided through the
                config.WhisperLanguages enum, which guarantees support, or as a string.

        Returns:
            The model's response.
        """
        if isinstance(language, config.WhisperLanguages):
            language = language.value

        return await self.client.audio.transcriptions.create(
            model=model,
            file=pathlib.Path(audio_file),
            response_format="text",
            language=language,
        )  # type: ignore[return-value] # response_format overrides output type.


class ImageGeneration(OpenAIBaseClass):
    """A class for running the Image Generation models."""

    async def run(  # noqa: PLR0913
        self,
        prompt: str,
        model: str = "dall-e-3",
        size: Literal["256x256", "512x512", "1024x1024", "1792x1024", "1024x1792"]
        | None = None,
        quality: Literal["standard", "hd"] = "standard",
        n: int | None = None,
    ) -> list[str | None]:
        """Runs the Image Generation model.

        Args:
            prompt: The prompt to generate an image from.
            model: The name of the Image Generation model to use.
            size: The size of the generated image.
            quality: The quality of the generated image.
            n: The number of images to generate.

        Returns:
            str: The image urls.
        """
        response = await self.client.images.generate(
            model=model,
            prompt=prompt,
            size=size,
            quality=quality,
            n=n,
        )

        return [data.url for data in response.data]


class Embedding(OpenAIBaseClass):
    """A class for running the Embedding models."""

    async def run(
        self,
        text: str,
        model: Literal[
            "text-embedding-3-small",
            "text-embedding-3-large",
        ] = "text-embedding-3-large",
        *,
        keep_new_lines: bool = False,
    ) -> list[float]:
        """Runs the Embedding model.

        Args:
            text: the string to embed.
            model: the name of the Embedding model to use.
            keep_new_lines: Whether to keep or remove line breaks,
            defaults to False.

        Returns:
            The embedding (list of numbers)
        """
        if keep_new_lines is False:
            text = text.replace("\n", " ")
        response = await self.client.embeddings.create(
            input=text,
            model=model,
        )
        return response.data[0].embedding



"""Custom exceptions."""
from cloai.core import config

logger = config.get_logger()


class LoggedException(BaseException):
    """Base class for exceptions that log their message."""

    def __init__(self, message: str) -> None:
        """Initializes a new instance of the LoggedException class.

        Args:
            message: The message to display.
        """
        logger.exception(message)
        super().__init__(message)


class FileSizeError(LoggedException):
    """Raised when the file is too large."""


class InvalidArgumentError(LoggedException):
    """Raised when the arguments are invalid."""


class LoggedValueError(LoggedException):
    """Raised when a value is invalid."""


class OpenAIError(LoggedException):
    """Raised when an OpenAI error occurs."""
